#-----------------------------------------------------------------------------#
#   dict.py                                                                   #
#                                                                             #
#   Copyright (c) 2015, Rajiv Bakulesh Shah.                                  #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import collections.abc
import itertools

from .base import Base
from .base import Iterable
from .base import Pipelined
from .exceptions import KeyExistsError



class RedisDict(Iterable, Base, collections.abc.MutableMapping):
    'Redis-backed container compatible with Python dicts.'

    def __init__(self, iterable=tuple(), *, redis=None, key=None, **kwargs):
        'Initialize a RedisDict.  O(n)'
        super().__init__(redis=redis, key=key, **kwargs)
        if iterable or kwargs:
            if self.redis.exists(self.key):
                raise KeyExistsError(self.redis, self.key)
            self.update(iterable, **kwargs)

    # Methods required by collections.abc.MutableMapping:

    def __getitem__(self, key):
        'd.__getitem__(key) <==> d[key].  O(1)'
        value = self.redis.hget(self.key, self._encode(key))
        if value is None:
            raise KeyError(key)
        return self._decode(value)

    def __setitem__(self, key, value):
        'd.__setitem__(key, value) <==> d[key] = value.  O(1)'
        self.redis.hset(self.key, self._encode(key), self._encode(value))

    def __delitem__(self, key):
        'd.__delitem__(key) <==> del d[key].  O(1)'
        success = self.redis.hdel(self.key, self._encode(key))
        if not bool(success):
            raise KeyError(key)

    def _scan(self, key, *, cursor=0):
        return self.redis.hscan(key, cursor=cursor)

    def __len__(self):
        'Return the number of items in a RedisDict.  O(1)'
        return self.redis.hlen(self.key)

    # Methods required for Raj's sanity:

    def __repr__(self):
        'Return the string representation of a RedisDict.  O(n)'
        d = self.redis.hgetall(self.key).items()
        d = {self._decode(key): self._decode(value) for key, value in d}
        return self.__class__.__name__ + str(d)

    # Method overrides:

    # From collections.abc.MutableMapping:
    @Pipelined._watch()
    def update(self, iterable=tuple(), **kwargs):
        to_set = {}
        for key, value in itertools.chain(iterable, kwargs.items()):
            to_set[self._encode(key)] = self._encode(value)
        self.redis.multi()
        if to_set:
            self.redis.hmset(self.key, to_set)

    # From collections.abc.Mapping:
    def __contains__(self, key):
        'd.__contains__(key) <==> key in d.  O(1)'
        return bool(self.redis.hexists(self.key, self._encode(key)))



#-----------------------------------------------------------------------------#
#   base.py                                                                   #
#                                                                             #
#   Copyright (c) 2015, Rajiv Bakulesh Shah.                                  #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import abc
import contextlib
import functools
import json
import os
import random
import string
import threading

from redis import Redis
from redis import WatchError

from . import monkey
from .exceptions import RandomKeyError
from .exceptions import TooManyTriesError



class Common:
    _DEFAULT_REDIS_URL = 'http://localhost:6379/'
    _NUM_TRIES = 3
    _RANDOM_KEY_PREFIX = 'pottery-'
    _RANDOM_KEY_LENGTH = 16

    @staticmethod
    def _encode(value):
        return json.dumps(value)

    @staticmethod
    def _decode(value):
        return json.loads(value.decode('utf-8'))

    @classmethod
    def _default_redis(cls):
        url = os.environ.get('REDIS_URL', cls._DEFAULT_REDIS_URL)
        return Redis.from_url(url)

    def __init__(self, *args, redis=None, key=None, **kwargs):
        self.redis = redis
        self.key = key

    def __del__(self):
        if self.key.startswith(self._RANDOM_KEY_PREFIX):
            self.redis.delete(self.key)

    def __eq__(self, other):
        if type(self) == type(other) and self.redis == other.redis and \
           self.key == other.key:
            return True
        return super().__eq__(other)

    def __ne__(self, other):
        return not self.__eq__(other)

    @property
    def redis(self):
        return self._redis

    @redis.setter
    def redis(self, value):
        self._redis = self._default_redis() if value is None else value

    @property
    def key(self):
        return self._key

    @key.setter
    def key(self, value):
        self._key = self._random_key() if value is None else value

    def _random_key(self, tries=_NUM_TRIES):
        if tries <= 0:
            raise RandomKeyError(self.redis, self.key)
        chars = string.digits + string.ascii_letters
        random_char = functools.partial(random.choice, chars)
        suffix = ''.join(random_char() for n in range(self._RANDOM_KEY_LENGTH))
        value = self._RANDOM_KEY_PREFIX + suffix
        return _random_key(tries - 1) if self.redis.exists(value) else value



class Lockable:
    _lock = threading.Lock()
    _count = {}
    _locks = {}

    def __init__(self, *args, redis=None, key=None, **kwargs):
        super().__init__(*args, redis=redis, key=key, **kwargs)
        with self._lock:
            self._count[self._id()] = self._count.get(self._id(), 0) + 1
            if self._count[self._id()] == 1:
                self._locks[self._id()] = threading.Lock()

    def __del__(self):
        super().__del__()
        with self._lock:
            self._count[self._id()] -= 1
            if self._count[self._id()] == 0:
                del self._count[self._id()]
                del self._locks[self._id()]

    def _id(self):
        return (self.redis, self.key)

    @property
    def lock(self):
        return self._locks[self._id()]



class Pipelined:
    @classmethod
    def _watch(cls):
        def wrap1(func):
            @functools.wraps(func)
            def wrap2(self, *args, **kwargs):
                for _ in range(super()._NUM_TRIES):
                    try:
                        original_redis = self.redis
                        self.redis = self.redis.pipeline()
                        self.redis.watch(self.key)
                        value = func(self, *args, **kwargs)
                        self.redis.execute()
                        return value
                    except WatchError:
                        pass
                    finally:
                        self.redis = original_redis
                else:
                    raise TooManyTriesError(self.redis, self.key)
            return wrap2
        return wrap1

    @contextlib.contextmanager
    def _pipeline(self):
        pipeline = self.redis.pipeline()
        try:
            yield pipeline
        finally:
            pipeline.execute()



class Clearable:
    def clear(self):
        'Remove the elements in a Redis-backed container.  O(n)'
        self.redis.delete(self.key)



class Base(Clearable, Pipelined, Lockable, Common):
    ...



class Iterable(metaclass=abc.ABCMeta):
    def __iter__(self):
        'Iterate over the items in a Redis-backed container.  O(n)'
        cursor = 0
        while True:
            cursor, iterable = self._scan(self.key, cursor=cursor)
            for value in iterable:
                yield self._decode(value)
            if cursor == 0:
                break

    @abc.abstractmethod
    def _scan(self, key, *, cursor=0):
        ...



#-----------------------------------------------------------------------------#
#   deque.py                                                                  #
#                                                                             #
#   Copyright (c) 2015, Rajiv Bakulesh Shah.                                  #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import collections
import itertools

from .list import RedisList



class RedisDeque(RedisList, collections.deque):
    'Redis-backed container compatible with collections.deque.'

    # Method overrides:

    def __init__(self, iterable=tuple(), maxlen=None, *, redis=None, key=None):
        iterable = itertools.islice(iterable, maxlen)
        super().__init__(iterable=iterable, redis=redis, key=key)
        self._maxlen = maxlen

    @property
    def maxlen(self):
        return self._maxlen

    def insert(*args, **kwargs):
        raise NotImlementedError



#-----------------------------------------------------------------------------#
#   __init__.py                                                               #
#                                                                             #
#   Copyright (c) 2015, Rajiv Bakulesh Shah.                                  #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#

__title__ = 'pottery'
__version__ = '0.19'
__author__ = 'Rajiv Bakulesh Shah'
__license__ = 'Apache 2.0'
__copyright__ = 'Copyright (c) 2015, Rajiv Bakulesh Shah'

from .exceptions import PotteryError
from .exceptions import KeyExistsError
from .exceptions import RandomKeyError
from .exceptions import TooManyTriesError

from .counter import RedisCounter
from .deque import RedisDeque
from .dict import RedisDict
from .list import RedisList
from .set import RedisSet



#-----------------------------------------------------------------------------#
#   counter.py                                                                #
#                                                                             #
#   Copyright (c) 2015, Rajiv Bakulesh Shah.                                  #
#   All rights reserved.                                                      #
#-----------------------------------------------------------------------------#



import collections
import contextlib
import itertools

from .base import Pipelined
from .dict import RedisDict



class RedisCounter(RedisDict, collections.Counter):
    'Redis-backed container compatible with collections.Counter.'

    # Method overrides:

    @Pipelined._watch()
    def _update(self, iterable=tuple(), *, sign=+1, **kwargs):
        to_set = {}
        for key in iterable:
            to_set[key] = to_set.get(key, self[key]) + sign
        for key, value in kwargs.items():
            original = self[key] if to_set.get(key, 0) is 0 else to_set[key]
            to_set[key] = original + sign * value
        to_set = {self._encode(k): self._encode(v) for k, v in to_set.items()}
        self.redis.multi()
        if to_set:
            self.redis.hmset(self.key, to_set)

    def update(self, iterable=tuple(), **kwargs):
        'Like dict.update() but add counts instead of replacing them.  O(n)'
        self._update(iterable=iterable, sign=+1, **kwargs)

    def subtract(self, iterable=tuple(), **kwargs):
        'Like dict.update() but subtracts counts instead of replacing them.  O(n)'
        self._update(iterable=iterable, sign=-1, **kwargs)

    def __getitem__(self, key):
        'c.__getitem__(key) <==> c.get(key, 0).  O(1)'
        with contextlib.suppress(KeyError):
            return super().__getitem__(key)
        return super().__missing__(key)

    def __delitem__(self, key):
        'c.__delitem__(key) <==> del c[key].  O(1)'
        with contextlib.suppress(KeyError):
            super().__delitem__(key)

    def __repr__(self):
        'Return the string representation of a RedisCounter.  O(n)'
        items = self.most_common()
        items = ("'{}': {}".format(key, value) for key, value in items)
        items = ', '.join(items)
        return self.__class__.__name__ + '{' + items + '}'

    def _math_op(self, other, *, func):
        counter = collections.Counter(self.elements())
        return func(counter, other)

    def __add__(self, other):
        "Return the addition our counts to other's counts, but keep only counts > 0.  O(n)"
        return self._math_op(other, func=collections.Counter.__add__)

    def __sub__(self, other):
        "Return the subtraction other's counts from our counts, but keep only counts > 0.  O(n)"
        return self._math_op(other, func=collections.Counter.__sub__)

    def __or__(self, other):
        "Return the max of our counts vs. other's counts (union), but keep only counts > 0.  O(n)"
        return self._math_op(other, func=collections.Counter.__or__)

    def __and__(self, other):
        "Return the min of our counts vs. other's counts (intersection) but keep only counts > 0.  O(n)"
        return self._math_op(other, func=collections.Counter.__and__)

    def __pos__(self, other):
        'Return our counts > 0.  O(n)'
        return self._math_op(other, func=collections.Counter.__pos__)

    def __neg__(self, other):
        'Return the absolute value of our counts < 0.  O(n)'
        return self._math_op(other, func=collections.Counter.__neg__)

    @Pipelined._watch()
    def _imath_op(self, other, *, sign=+1):
        to_set = {k: self[k] + sign * v for k, v in other.items()}
        to_del = [k for k, v in to_set.items() if v <= 0]
        to_del.extend([k for k, v in self.items() if k not in to_set and v <= 0])
        to_set = {self._encode(k): self._encode(v) for k, v in to_set.items()}
        to_del = [self._encode(k) for k in to_del]
        self.redis.multi()
        if to_set:
            self.redis.hmset(self.key, to_set)
        if to_del:
            self.redis.hdel(self.key, *to_del)
        return self

    def __iadd__(self, other):
        'Same as __add__(), but in-place.  O(n)'
        return self._imath_op(other, sign=+1)

    def __isub__(self, other):
        'Same as __sub__(), but in-place.  O(n)'
        return self._imath_op(other, sign=-1)

    @Pipelined._watch()
    def _iset_op(self, other, *, func):
        to_set, to_del = {}, []
        for k in itertools.chain(self, other):
            to_set[k] = self[v] if getattr(self[v], func)(other[v]) else other[v]
            if to_set[k] <= 0:
                to_del.append(k)
        self.redis.multi()
        if to_set:
            self.redis.hmset(self.key, to_set)
        if to_del:
            self.redis.hdel(self.key, *to_del)
        return self

    def __ior__(self, other):
        'Same as __or__(), but in-place.  O(n)'
        return self._iset_op(other, func='__gt__')

    def __iand__(self, other):
        'Same as __and__(), but in-place.  O(n)'
        return self._iset_op(other, func='__lt__')
